from typing import Any, Dict, List, NamedTuple, Optional, Union

from graphql import (
    get_named_type,
    resolve_thunk,
    GraphQLArgument,
    GraphQLArgumentMap,
    GraphQLBoolean,
    GraphQLField,
    GraphQLFieldResolver,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLString,
    ThunkMapping,
)

from graphql import GraphQLNamedOutputType

try:
    from typing import Protocol
except ImportError:  # Python < 3.8
    from typing_extensions import Protocol  # type: ignore

__all__ = [
    "backward_connection_args",
    "connection_args",
    "connection_definitions",
    "forward_connection_args",
    "page_info_type",
    "Connection",
    "ConnectionArguments",
    "ConnectionConstructor",
    "ConnectionCursor",
    "ConnectionType",
    "Edge",
    "EdgeConstructor",
    "EdgeType",
    "GraphQLConnectionDefinitions",
    "PageInfo",
    "PageInfoConstructor",
    "PageInfoType",
]


# Returns a GraphQLArgumentMap appropriate to include on a field
# whose return type is a connection type with forward pagination.
forward_connection_args: GraphQLArgumentMap = {
    "after": GraphQLArgument(
        GraphQLString,
        description="Returns the items in the list"
        " that come after the specified cursor.",
    ),
    "first": GraphQLArgument(
        GraphQLInt,
        description="Returns the first n items from the list.",
    ),
}

# Returns a GraphQLArgumentMap appropriate to include on a field
# whose return type is a connection type with backward pagination.
backward_connection_args: GraphQLArgumentMap = {
    "before": GraphQLArgument(
        GraphQLString,
        description="Returns the items in the list"
        " that come before the specified cursor.",
    ),
    "last": GraphQLArgument(
        GraphQLInt, description="Returns the last n items from the list."
    ),
}

# Returns a GraphQLArgumentMap appropriate to include on a field
# whose return type is a connection type with bidirectional pagination.
connection_args = {**forward_connection_args, **backward_connection_args}


class GraphQLConnectionDefinitions(NamedTuple):
    edge_type: GraphQLObjectType
    connection_type: GraphQLObjectType


"""A type alias for cursors in this implementation."""
ConnectionCursor = str


"""A type describing the arguments a connection field receives in GraphQL.

The following kinds of arguments are expected (all optional):

    before: ConnectionCursor
    after: ConnectionCursor
    first: int
    last: int
"""
ConnectionArguments = Dict[str, Any]


def connection_definitions(
    node_type: Union[GraphQLNamedOutputType, GraphQLNonNull[GraphQLNamedOutputType]],
    name: Optional[str] = None,
    resolve_node: Optional[GraphQLFieldResolver] = None,
    resolve_cursor: Optional[GraphQLFieldResolver] = None,
    edge_fields: Optional[ThunkMapping[GraphQLField]] = None,
    connection_fields: Optional[ThunkMapping[GraphQLField]] = None,
) -> GraphQLConnectionDefinitions:
    """Return GraphQLObjectTypes for a connection with the given name.

    The nodes of the returned object types will be of the specified type.
    """
    name = name or get_named_type(node_type).name

    edge_type = GraphQLObjectType(
        name + "Edge",
        description="An edge in a connection.",
        fields=lambda: {
            "node": GraphQLField(
                node_type,
                resolve=resolve_node,
                description="The item at the end of the edge",
            ),
            "cursor": GraphQLField(
                GraphQLNonNull(GraphQLString),
                resolve=resolve_cursor,
                description="A cursor for use in pagination",
            ),
            **resolve_thunk(edge_fields or {}),
        },
    )

    connection_type = GraphQLObjectType(
        name + "Connection",
        description="A connection to a list of items.",
        fields=lambda: {
            "pageInfo": GraphQLField(
                GraphQLNonNull(page_info_type),
                description="Information to aid in pagination.",
            ),
            "edges": GraphQLField(
                GraphQLList(edge_type), description="A list of edges."
            ),
            **resolve_thunk(connection_fields or {}),
        },
    )

    return GraphQLConnectionDefinitions(edge_type, connection_type)


class PageInfoType(Protocol):
    @property
    def startCursor(self) -> Optional[ConnectionCursor]:
        ...

    def endCursor(self) -> Optional[ConnectionCursor]:
        ...

    def hasPreviousPage(self) -> bool:
        ...

    def hasNextPage(self) -> bool:
        ...


class PageInfoConstructor(Protocol):
    def __call__(
        self,
        *,
        startCursor: Optional[ConnectionCursor],
        endCursor: Optional[ConnectionCursor],
        hasPreviousPage: bool,
        hasNextPage: bool,
    ) -> PageInfoType:
        ...


class PageInfo(NamedTuple):
    """A type designed to be exposed as `PageInfo` over GraphQL."""

    startCursor: Optional[ConnectionCursor]
    endCursor: Optional[ConnectionCursor]
    hasPreviousPage: bool
    hasNextPage: bool


class EdgeType(Protocol):
    @property
    def node(self) -> Any:
        ...

    @property
    def cursor(self) -> ConnectionCursor:
        ...


class EdgeConstructor(Protocol):
    def __call__(self, *, node: Any, cursor: ConnectionCursor) -> EdgeType:
        ...


class Edge(NamedTuple):
    """A type designed to be exposed as a `Edge` over GraphQL."""

    node: Any
    cursor: ConnectionCursor


class ConnectionType(Protocol):
    @property
    def edges(self) -> List[EdgeType]:
        ...

    @property
    def pageInfo(self) -> PageInfoType:
        ...


class ConnectionConstructor(Protocol):
    def __call__(
        self,
        *,
        edges: List[EdgeType],
        pageInfo: PageInfoType,
    ) -> ConnectionType:
        ...


class Connection(NamedTuple):
    """A type designed to be exposed as a `Connection` over GraphQL."""

    edges: List[Edge]
    pageInfo: PageInfo


# The common page info type used by all connections.
page_info_type = GraphQLObjectType(
    "PageInfo",
    description="Information about pagination in a connection.",
    fields=lambda: {
        "hasNextPage": GraphQLField(
            GraphQLNonNull(GraphQLBoolean),
            description="When paginating forwards, are there more items?",
        ),
        "hasPreviousPage": GraphQLField(
            GraphQLNonNull(GraphQLBoolean),
            description="When paginating backwards, are there more items?",
        ),
        "startCursor": GraphQLField(
            GraphQLString,
            description="When paginating backwards, the cursor to continue.",
        ),
        "endCursor": GraphQLField(
            GraphQLString,
            description="When paginating forwards, the cursor to continue.",
        ),
    },
)



from graphql import graphql_sync

from .star_wars_schema import star_wars_schema as schema


def describe_star_wars_mutations():
    def correctly_mutates_dataset():
        source = """
          mutation ($input: IntroduceShipInput!) {
            introduceShip(input: $input) {
              ship {
                id
                name
              }
              faction {
                name
              }
              clientMutationId
            }
          }
        """
        variable_values = {
            "input": {
                "shipName": "B-Wing",
                "factionId": "1",
                "clientMutationId": "abcde",
            }
        }
        expected = {
            "introduceShip": {
                "ship": {"id": "U2hpcDo5", "name": "B-Wing"},
                "faction": {"name": "Alliance to Restore the Republic"},
                "clientMutationId": "abcde",
            }
        }
        result = graphql_sync(schema, source, variable_values=variable_values)
        assert result == (expected, None)



from textwrap import dedent as _dedent

__all__ = ["dedent"]


def dedent(text: str) -> str:
    """Fix indentation and also trim given text string."""
    return _dedent(text.lstrip("\n").rstrip(" \t\n"))



from graphql_relay.utils import base64, unbase64


example_unicode = "Some examples:  ͢❤😀"
example_base64 = "U29tZSBleGFtcGxlczogIM2i4p2k8J+YgA=="


def describe_base64_conversion():
    def converts_from_unicode_to_base64():
        assert base64(example_unicode) == example_base64

    def converts_from_base64_to_unicode():
        assert unbase64(example_base64) == example_unicode

    def converts_invalid_base64_to_empty_string():
        assert unbase64("") == ""
        assert unbase64("invalid") == ""
        assert unbase64(example_base64[-1:]) == ""
        assert unbase64(example_base64[1:]) == ""
        assert unbase64("!" + example_base64[1:]) == ""
        assert unbase64("Ü" + example_base64[1:]) == ""

    def converts_from_unicode_as_bytes_to_base64():
        bytes_example_code = example_unicode.encode("utf-8")
        assert base64(bytes_example_code) == example_base64  # type: ignore
        bytearray_example_code = bytearray(bytes_example_code)
        assert base64(bytearray_example_code) == example_base64  # type: ignore

    def converts_from_base64_as_bytes_to_unicode():
        bytes_example_code = example_base64.encode("ascii")
        assert unbase64(bytes_example_code) == example_unicode  # type: ignore
        bytearray_example_code = bytearray(bytes_example_code)
        assert unbase64(bytearray_example_code) == example_unicode  # type: ignore



from typing import Any, NamedTuple, Optional

from pytest import fixture

from graphql import (
    graphql_sync,
    GraphQLField,
    GraphQLList,
    GraphQLInt,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
    GraphQLString,
)

from graphql_relay import from_global_id, global_id_field, node_definitions


class User(NamedTuple):
    id: str
    name: str


class Photo(NamedTuple):
    photo_id: str
    width: int


class Post(NamedTuple):
    id: str
    text: str


@fixture(scope="module", params=["object_access", "dict_access"])
def schema(request):
    """Run each test with object access and dict access."""
    use_dicts = request.param == "dict_access"

    user_cls = dict if use_dicts else User
    user_data = [
        user_cls(id="1", name="John Doe"),
        user_cls(id="2", name="Jane Smith"),
    ]

    photo_cls = dict if use_dicts else Photo
    photo_data = [
        photo_cls(photo_id="1", width=300),
        photo_cls(photo_id="2", width=400),
    ]

    post_cls = dict if use_dicts else Post
    post_data = [post_cls(id="1", text="lorem"), post_cls(id="2", text="ipsum")]

    if use_dicts:

        def get_node(global_id: str, info: GraphQLResolveInfo) -> Any:
            assert info.schema is schema
            type_, id_ = from_global_id(global_id)
            if type_ == "User":
                return next(filter(lambda obj: obj["id"] == id_, user_data), None)
            if type_ == "Photo":
                return next(
                    filter(lambda obj: obj["photo_id"] == id_, photo_data), None
                )
            if type_ == "Post":
                return next(filter(lambda obj: obj["id"] == id_, post_data), None)
            return None  # pragma: no cover

        def get_node_type(
            obj: Any, info: GraphQLResolveInfo, _type: Any
        ) -> Optional[str]:
            assert info.schema is schema
            if "name" in obj:
                return user_type.name
            if "photo_id" in obj:
                return photo_type.name
            if "text" in obj:
                return post_type.name
            return None  # pragma: no cover

    else:

        def get_node(global_id: str, info: GraphQLResolveInfo) -> Any:
            assert info.schema is schema
            type_, id_ = from_global_id(global_id)
            if type_ == "User":
                return next(filter(lambda obj: obj.id == id_, user_data), None)
            if type_ == "Photo":
                return next(filter(lambda obj: obj.photo_id == id_, photo_data), None)
            if type_ == "Post":
                return next(filter(lambda obj: obj.id == id_, post_data), None)
            return None  # pragma: no cover

        def get_node_type(
            obj: Any, info: GraphQLResolveInfo, _type: Any
        ) -> Optional[str]:
            assert info.schema is schema
            if isinstance(obj, User):
                return user_type.name
            if isinstance(obj, Photo):
                return photo_type.name
            if isinstance(obj, Post):
                return post_type.name
            return None  # pragma: no cover

    node_interface, node_field = node_definitions(get_node, get_node_type)[:2]

    user_type = GraphQLObjectType(
        "User",
        fields=lambda: {
            "id": global_id_field("User"),
            "name": GraphQLField(GraphQLString),
        },
        interfaces=[node_interface],
    )

    photo_type = GraphQLObjectType(
        "Photo",
        fields=lambda: {
            "id": global_id_field(
                "Photo",
                lambda obj, _info: obj["photo_id"] if use_dicts else obj.photo_id,
            ),
            "width": GraphQLField(GraphQLInt),
        },
        interfaces=[node_interface],
    )

    post_type = GraphQLObjectType(
        "Post",
        fields=lambda: {"id": global_id_field(), "text": GraphQLField(GraphQLString)},
        interfaces=[node_interface],
    )

    query_type = GraphQLObjectType(
        "Query",
        fields=lambda: {
            "node": node_field,
            "allObjects": GraphQLField(
                GraphQLList(node_interface),
                resolve=lambda _root, _info: [*user_data, *photo_data, *post_data],
            ),
        },
    )

    schema = GraphQLSchema(query=query_type, types=[user_type, photo_type, post_type])

    yield schema


def describe_global_id_fields():
    def gives_different_ids(schema):
        source = """
        {
          allObjects {
            id
          }
        }
        """
        assert graphql_sync(schema, source) == (
            {
                "allObjects": [
                    {"id": "VXNlcjox"},
                    {"id": "VXNlcjoy"},
                    {"id": "UGhvdG86MQ=="},
                    {"id": "UGhvdG86Mg=="},
                    {"id": "UG9zdDox"},
                    {"id": "UG9zdDoy"},
                ]
            },
            None,
        )

    def allows_to_refetch_the_ids(schema):
        source = """
        {
          user: node(id: "VXNlcjox") {
            id
            ... on User {
              name
            }
          },
          photo: node(id: "UGhvdG86MQ==") {
            id
            ... on Photo {
              width
            }
          }
          post: node(id: "UG9zdDox") {
            id
            ... on Post {
              text
            }
          }
        }
        """
        assert graphql_sync(schema, source) == (
            {
                "user": {"id": "VXNlcjox", "name": "John Doe"},
                "photo": {"id": "UGhvdG86MQ==", "width": 300},
                "post": {"id": "UG9zdDox", "text": "lorem"},
            },
            None,
        )

    def handles_valid_global_ids():
        assert from_global_id("Zm9v") == ("", "foo")
        assert from_global_id(b"Zm9v") == ("", "foo")  # type: ignore
        assert from_global_id("Zm9vOmJhcg==") == ("foo", "bar")
        assert from_global_id(b"Zm9vOmJhcg==") == ("foo", "bar")  # type: ignore

    def handles_invalid_global_ids():
        assert from_global_id("") == ("", "")
        assert from_global_id("Og==") == ("", "")
        assert from_global_id("bad!") == ("", "")
        assert from_global_id("invalid") == ("", "")



from typing import Any, Iterator, Optional, Sequence

try:
    from typing import Protocol
except ImportError:  # Python < 3.8
    from typing_extensions import Protocol  # type: ignore

from ..utils.base64 import base64, unbase64
from .connection import (
    Connection,
    ConnectionArguments,
    ConnectionConstructor,
    ConnectionCursor,
    ConnectionType,
    Edge,
    EdgeConstructor,
    PageInfo,
    PageInfoConstructor,
)

__all__ = [
    "connection_from_array",
    "connection_from_array_slice",
    "cursor_for_object_in_connection",
    "cursor_to_offset",
    "get_offset_with_default",
    "offset_to_cursor",
    "SizedSliceable",
]


class SizedSliceable(Protocol):
    def __getitem__(self, index: slice) -> Any:
        ...

    def __iter__(self) -> Iterator:
        ...

    def __len__(self) -> int:
        ...


def connection_from_array(
    data: SizedSliceable,
    args: Optional[ConnectionArguments] = None,
    connection_type: ConnectionConstructor = Connection,
    edge_type: EdgeConstructor = Edge,
    page_info_type: PageInfoConstructor = PageInfo,
) -> ConnectionType:
    """Create a connection object from a sequence of objects.

    Note that different from its JavaScript counterpart which expects an array,
    this function accepts any kind of sliceable object with a length.

    Given this `data` object representing the result set, and connection arguments,
    this simple function returns a connection object for use in GraphQL. It uses
    offsets as pagination, so pagination will only work if the data is static.

    The result will use the default types provided in the `connectiontypes` module
    if you don't pass custom types as arguments.
    """
    return connection_from_array_slice(
        data,
        args,
        slice_start=0,
        array_length=len(data),
        connection_type=connection_type,
        edge_type=edge_type,
        page_info_type=page_info_type,
    )


def connection_from_array_slice(
    array_slice: SizedSliceable,
    args: Optional[ConnectionArguments] = None,
    slice_start: int = 0,
    array_length: Optional[int] = None,
    array_slice_length: Optional[int] = None,
    connection_type: ConnectionConstructor = Connection,
    edge_type: EdgeConstructor = Edge,
    page_info_type: PageInfoConstructor = PageInfo,
) -> ConnectionType:
    """Create a connection object from a slice of the result set.

    Note that different from its JavaScript counterpart which expects an array,
    this function accepts any kind of sliceable object. This object represents
    a slice of the full result set. You need to pass the start position of the
    slice as `slice start` and the length of the full result set as `array_length`.
    If the `array_slice` does not have a length, you need to provide it separately
    in `array_slice_length` as well.

    This function is similar to `connection_from_array`, but is intended for use
    cases where you know the cardinality of the connection, consider it too large
    to materialize the entire result set, and instead wish to pass in only a slice
    of the total result large enough to cover the range specified in `args`.

    If you do not provide a `slice_start`, we assume that the slice starts at
    the beginning of the result set, and if you do not provide an `array_length`,
    we assume that the slice ends at the end of the result set.
    """
    args = args or {}
    before = args.get("before")
    after = args.get("after")
    first = args.get("first")
    last = args.get("last")
    if array_slice_length is None:
        array_slice_length = len(array_slice)
    slice_end = slice_start + array_slice_length
    if array_length is None:
        array_length = slice_end

    start_offset = max(slice_start, 0)
    end_offset = min(slice_end, array_length)

    after_offset = get_offset_with_default(after, -1)
    if 0 <= after_offset < array_length:
        start_offset = max(start_offset, after_offset + 1)

    before_offset = get_offset_with_default(before, end_offset)
    if 0 <= before_offset < array_length:
        end_offset = min(end_offset, before_offset)

    if isinstance(first, int):
        if first < 0:
            raise ValueError("Argument 'first' must be a non-negative integer.")

        end_offset = min(end_offset, start_offset + first)
    if isinstance(last, int):
        if last < 0:
            raise ValueError("Argument 'last' must be a non-negative integer.")

        start_offset = max(start_offset, end_offset - last)

    # If supplied slice is too large, trim it down before mapping over it.
    trimmed_slice = array_slice[start_offset - slice_start : end_offset - slice_start]

    edges = [
        edge_type(node=value, cursor=offset_to_cursor(start_offset + index))
        for index, value in enumerate(trimmed_slice)
    ]

    first_edge_cursor = edges[0].cursor if edges else None
    last_edge_cursor = edges[-1].cursor if edges else None
    lower_bound = after_offset + 1 if after else 0
    upper_bound = before_offset if before else array_length

    return connection_type(
        edges=edges,
        pageInfo=page_info_type(
            startCursor=first_edge_cursor,
            endCursor=last_edge_cursor,
            hasPreviousPage=isinstance(last, int) and start_offset > lower_bound,
            hasNextPage=isinstance(first, int) and end_offset < upper_bound,
        ),
    )


PREFIX = "arrayconnection:"


def offset_to_cursor(offset: int) -> ConnectionCursor:
    """Create the cursor string from an offset."""
    return base64(f"{PREFIX}{offset}")


def cursor_to_offset(cursor: ConnectionCursor) -> Optional[int]:
    """Extract the offset from the cursor string."""
    try:
        return int(unbase64(cursor)[len(PREFIX) :])
    except ValueError:
        return None


def cursor_for_object_in_connection(
    data: Sequence, obj: Any
) -> Optional[ConnectionCursor]:
    """Return the cursor associated with an object in a sequence.

    This function uses the `index` method of the sequence if it exists,
    otherwise searches the object by iterating via the `__getitem__` method.
    """
    try:
        offset = data.index(obj)
    except AttributeError:
        # data does not have an index method
        offset = 0
        try:
            while True:
                if data[offset] == obj:
                    break
                offset += 1
        except IndexError:
            return None
        else:
            return offset_to_cursor(offset)
    except ValueError:
        return None
    else:
        return offset_to_cursor(offset)


def get_offset_with_default(
    cursor: Optional[ConnectionCursor] = None, default_offset: int = 0
) -> int:
    """Get offset from a given cursor and a default.

    Given an optional cursor and a default offset, return the offset to use;
    if the cursor contains a valid offset, that will be used,
    otherwise it will be the default.
    """
    if not isinstance(cursor, str):
        return default_offset

    offset = cursor_to_offset(cursor)
    return default_offset if offset is None else offset



from base64 import b64encode, b64decode
import binascii

__all__ = ["base64", "unbase64"]

Base64String = str


def base64(s: str) -> Base64String:
    """Encode the string s using Base64."""
    b: bytes = s.encode("utf-8") if isinstance(s, str) else s
    return b64encode(b).decode("ascii")


def unbase64(s: Base64String) -> str:
    """Decode the string s using Base64."""
    try:
        b: bytes = s.encode("ascii") if isinstance(s, str) else s
    except UnicodeEncodeError:
        return ""
    try:
        return b64decode(b).decode("utf-8")
    except (binascii.Error, UnicodeDecodeError):
        return ""



from typing import Any, Callable, NamedTuple, Optional, Union

from graphql_relay.utils.base64 import base64, unbase64

from graphql import (
    GraphQLArgument,
    GraphQLNonNull,
    GraphQLID,
    GraphQLField,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLResolveInfo,
    GraphQLTypeResolver,
)

__all__ = [
    "from_global_id",
    "global_id_field",
    "node_definitions",
    "to_global_id",
    "GraphQLNodeDefinitions",
    "ResolvedGlobalId",
]


class GraphQLNodeDefinitions(NamedTuple):

    node_interface: GraphQLInterfaceType
    node_field: GraphQLField
    nodes_field: GraphQLField


def node_definitions(
    fetch_by_id: Callable[[str, GraphQLResolveInfo], Any],
    type_resolver: Optional[GraphQLTypeResolver] = None,
) -> GraphQLNodeDefinitions:
    """
    Given a function to map from an ID to an underlying object, and a function
    to map from an underlying object to the concrete GraphQLObjectType it
    corresponds to, constructs a `Node` interface that objects can implement,
    and a field object to be used as a `node` root field.

    If the type_resolver is omitted, object resolution on the interface will be
    handled with the `is_type_of` method on object types, as with any GraphQL
    interface without a provided `resolve_type` method.
    """
    node_interface = GraphQLInterfaceType(
        "Node",
        description="An object with an ID",
        fields=lambda: {
            "id": GraphQLField(
                GraphQLNonNull(GraphQLID), description="The id of the object."
            )
        },
        resolve_type=type_resolver,
    )

    # noinspection PyShadowingBuiltins
    node_field = GraphQLField(
        node_interface,
        description="Fetches an object given its ID",
        args={
            "id": GraphQLArgument(
                GraphQLNonNull(GraphQLID), description="The ID of an object"
            )
        },
        resolve=lambda _obj, info, id: fetch_by_id(id, info),
    )

    nodes_field = GraphQLField(
        GraphQLNonNull(GraphQLList(node_interface)),
        description="Fetches objects given their IDs",
        args={
            "ids": GraphQLArgument(
                GraphQLNonNull(GraphQLList(GraphQLNonNull(GraphQLID))),
                description="The IDs of objects",
            )
        },
        resolve=lambda _obj, info, ids: [fetch_by_id(id_, info) for id_ in ids],
    )

    return GraphQLNodeDefinitions(node_interface, node_field, nodes_field)


class ResolvedGlobalId(NamedTuple):

    type: str
    id: str


def to_global_id(type_: str, id_: Union[str, int]) -> str:
    """
    Takes a type name and an ID specific to that type name, and returns a
    "global ID" that is unique among all types.
    """
    return base64(f"{type_}:{GraphQLID.serialize(id_)}")


def from_global_id(global_id: str) -> ResolvedGlobalId:
    """
    Takes the "global ID" created by to_global_id, and returns the type name and ID
    used to create it.
    """
    global_id = unbase64(global_id)
    if ":" not in global_id:
        return ResolvedGlobalId("", global_id)
    return ResolvedGlobalId(*global_id.split(":", 1))


def global_id_field(
    type_name: Optional[str] = None,
    id_fetcher: Optional[Callable[[Any, GraphQLResolveInfo], str]] = None,
) -> GraphQLField:
    """
    Creates the configuration for an id field on a node, using `to_global_id` to
    construct the ID from the provided typename. The type-specific ID is fetched
    by calling id_fetcher on the object, or if not provided, by accessing the `id`
    attribute of the object, or the `id` if the object is a dict.
    """

    def resolve(obj: Any, info: GraphQLResolveInfo, **_args: Any) -> str:
        type_ = type_name or info.parent_type.name
        id_ = (
            id_fetcher(obj, info)
            if id_fetcher
            else (obj["id"] if isinstance(obj, dict) else obj.id)
        )
        return to_global_id(type_, id_)

    return GraphQLField(
        GraphQLNonNull(GraphQLID), description="The ID of an object", resolve=resolve
    )



import re

import graphql_relay
from graphql_relay.version import (
    VersionInfo,
    version,
    version_info,
    version_js,
    version_info_js,
)

_re_version = re.compile(r"(\d+)\.(\d+)\.(\d+)(?:(a|b|r?c)(\d+))?$")


def describe_version():
    def describe_version_info_class():
        def create_version_info_from_fields():
            v = VersionInfo(1, 2, 3, "alpha", 4)
            assert v.major == 1
            assert v.minor == 2
            assert v.micro == 3
            assert v.releaselevel == "alpha"
            assert v.serial == 4

        def create_version_info_from_str():
            v = VersionInfo.from_str("1.2.3")
            assert v.major == 1
            assert v.minor == 2
            assert v.micro == 3
            assert v.releaselevel == "final"
            assert v.serial == 0
            v = VersionInfo.from_str("1.2.3a4")
            assert v.major == 1
            assert v.minor == 2
            assert v.micro == 3
            assert v.releaselevel == "alpha"
            assert v.serial == 4
            v = VersionInfo.from_str("1.2.3beta4")
            assert v.major == 1
            assert v.minor == 2
            assert v.micro == 3
            assert v.releaselevel == "beta"
            assert v.serial == 4
            v = VersionInfo.from_str("12.34.56rc789")
            assert v.major == 12
            assert v.minor == 34
            assert v.micro == 56
            assert v.releaselevel == "candidate"
            assert v.serial == 789

        def serialize_as_str():
            v = VersionInfo(1, 2, 3, "final", 0)
            assert str(v) == "1.2.3"
            v = VersionInfo(1, 2, 3, "alpha", 4)
            assert str(v) == "1.2.3a4"
            v = VersionInfo(1, 2, 3, "candidate", 4)
            assert str(v) == "1.2.3rc4"

    def describe_graphql_core_version():
        def base_package_has_correct_version():
            assert graphql_relay.__version__ == version
            assert graphql_relay.version == version

        def base_package_has_correct_version_info():
            assert graphql_relay.__version_info__ is version_info
            assert graphql_relay.version_info is version_info

        def version_has_correct_format():
            assert isinstance(version, str)
            assert _re_version.match(version)

        def version_info_has_correct_fields():
            assert isinstance(version_info, tuple)
            assert str(version_info) == version
            groups = _re_version.match(version).groups()  # type: ignore
            assert version_info.major == int(groups[0])
            assert version_info.minor == int(groups[1])
            assert version_info.micro == int(groups[2])
            if groups[3] is None:  # pragma: no cover
                assert groups[4] is None
            else:  # pragma: no cover
                assert version_info.releaselevel[:1] == groups[3].lstrip("r")
                assert version_info.serial == int(groups[4])

    def describe_graphql_js_version():
        def base_package_has_correct_version_js():
            assert graphql_relay.__version_js__ == version_js
            assert graphql_relay.version_js == version_js

        def base_package_has_correct_version_info_js():
            assert graphql_relay.__version_info_js__ is version_info_js
            assert graphql_relay.version_info_js is version_info_js

        def version_js_has_correct_format():
            assert isinstance(version_js, str)
            assert _re_version.match(version_js)

        def version_info_js_has_correct_fields():
            assert isinstance(version_info_js, tuple)
            assert str(version_info_js) == version_js
            groups = _re_version.match(version_js).groups()  # type: ignore
            assert version_info_js.major == int(groups[0])
            assert version_info_js.minor == int(groups[1])
            assert version_info_js.micro == int(groups[2])
            if groups[3] is None:  # pragma: no cover
                assert groups[4] is None
            else:  # pragma: no cover
                assert version_info_js.releaselevel[:1] == groups[3].lstrip("r")
                assert version_info_js.serial == int(groups[4])



from graphql import (
    GraphQLID,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLInputField,
    GraphQLSchema,
    GraphQLString,
    GraphQLField,
)

from graphql_relay.node.node import node_definitions, global_id_field, from_global_id
from graphql_relay.connection.array_connection import connection_from_array
from graphql_relay.connection.connection import connection_args, connection_definitions
from graphql_relay.mutation.mutation import mutation_with_client_mutation_id

from .star_wars_data import (
    Faction,
    get_faction,
    get_ship,
    get_rebels,
    get_empire,
    create_ship,
)

# This is a basic end-to-end test, designed to demonstrate the various
# capabilities of a Relay-compliant GraphQL server.
#
# It is recommended that readers of this test be familiar with
# the end-to-end test in GraphQL.js first, as this test skips
# over the basics covered there in favor of illustrating the
# key aspects of the Relay spec that this test is designed to illustrate.
#
# We will create a GraphQL schema that describes the major
# factions and ships in the original Star Wars trilogy.
#
# NOTE: This may contain spoilers for the original Star
# Wars trilogy.

# Using our shorthand to describe type systems, the type system for our
# example will be the following:
#
# interface Node {
#   id: ID!
# }
#
# type Faction : Node {
#   id: ID!
#   name: String
#   ships: ShipConnection
# }
#
# type Ship : Node {
#   id: ID!
#   name: String
# }
#
# type ShipConnection {
#   edges: [ShipEdge]
#   pageInfo: PageInfo!
# }
#
# type ShipEdge {
#   cursor: String!
#   node: Ship
# }
#
# type PageInfo {
#   hasNextPage: Boolean!
#   hasPreviousPage: Boolean!
#   startCursor: String
#   endCursor: String
# }
#
# type Query {
#   rebels: Faction
#   empire: Faction
#   node(id: ID!): Node
# }
#
# input IntroduceShipInput {
#   clientMutationId: string
#   shipName: string!
#   factionId: ID!
# }
#
# type IntroduceShipPayload {
#   clientMutationId: string
#   ship: Ship
#   faction: Faction
# }
#
# type Mutation {
#   introduceShip(input IntroduceShipInput!): IntroduceShipPayload
# }

# We get the node interface and field from the relay library.
#
# The first method is the way we resolve an ID to its object. The second is the
# way we resolve an object that implements node to its type.


def get_node(global_id, _info):
    type_, id_ = from_global_id(global_id)
    if type_ == "Faction":
        return get_faction(id_)
    if type_ == "Ship":
        return get_ship(id_)
    return None  # pragma: no cover


def get_node_type(obj, _info, _type):
    if isinstance(obj, Faction):
        return faction_type.name
    return ship_type.name


node_interface, node_field = node_definitions(get_node, get_node_type)[:2]


# We define our basic ship type.
#
# This implements the following type system shorthand:
#   type Ship : Node {
#     id: String!
#     name: String
#   }
ship_type = GraphQLObjectType(
    name="Ship",
    description="A ship in the Star Wars saga",
    fields=lambda: {
        "id": global_id_field("Ship"),
        "name": GraphQLField(GraphQLString, description="The name of the ship."),
    },
    interfaces=[node_interface],
)

# We define a connection between a faction and its ships.
#
# connection_type implements the following type system shorthand:
#   type ShipConnection {
#     edges: [ShipEdge]
#     pageInfo: PageInfo!
#   }
#
# connection_type has an edges field - a list of edgeTypes that implement the
# following type system shorthand:
#   type ShipEdge {
#     cursor: String!
#     node: Ship
#   }
ship_edge, ship_connection = connection_definitions(ship_type, "Ship")

# We define our faction type, which implements the node interface.
#
# This implements the following type system shorthand:
#   type Faction : Node {
#     id: String!
#     name: String
#     ships: ShipConnection
#   }
faction_type = GraphQLObjectType(
    name="Faction",
    description="A faction in the Star Wars saga",
    fields=lambda: {
        "id": global_id_field("Faction"),
        "name": GraphQLField(GraphQLString, description="The name of the faction."),
        "ships": GraphQLField(
            ship_connection,
            description="The ships used by the faction.",
            args=connection_args,
            resolve=lambda faction, _info, **args: connection_from_array(
                [get_ship(ship) for ship in faction.ships], args
            ),
        ),
    },
    interfaces=[node_interface],
)

# This is the type that will be the root of our query, and the
# entry point into our schema.
#
# This implements the following type system shorthand:
#   type Query {
#     rebels: Faction
#     empire: Faction
#     node(id: String!): Node
#   }
query_type = GraphQLObjectType(
    name="Query",
    fields=lambda: {
        "rebels": GraphQLField(faction_type, resolve=lambda _obj, _info: get_rebels()),
        "empire": GraphQLField(faction_type, resolve=lambda _obj, _info: get_empire()),
        "node": node_field,
    },
)

# This will return a GraphQLFieldConfig for our ship
# mutation.
#
# It creates these two types implicitly:
#   input IntroduceShipInput {
#     clientMutationId: string
#     shipName: string!
#     factionId: ID!
#   }
#
#   type IntroduceShipPayload {
#     clientMutationId: string
#     ship: Ship
#     faction: Faction
#   }


class IntroduceShipMutation:

    # noinspection PyPep8Naming
    def __init__(self, shipId, factionId, clientMutationId=None):
        self.shipId = shipId
        self.factionId = factionId
        self.clientMutationId = clientMutationId


# noinspection PyPep8Naming
def mutate_and_get_payload(_info, shipName, factionId, **_input):
    new_ship = create_ship(shipName, factionId)
    return IntroduceShipMutation(shipId=new_ship.id, factionId=factionId)


ship_mutation = mutation_with_client_mutation_id(
    "IntroduceShip",
    input_fields={
        "shipName": GraphQLInputField(GraphQLNonNull(GraphQLString)),
        "factionId": GraphQLInputField(GraphQLNonNull(GraphQLID)),
    },
    output_fields={
        "ship": GraphQLField(
            ship_type, resolve=lambda payload, _info: get_ship(payload.shipId)
        ),
        "faction": GraphQLField(
            faction_type, resolve=lambda payload, _info: get_faction(payload.factionId)
        ),
    },
    mutate_and_get_payload=mutate_and_get_payload,
)

# This is the type that will be the root of our mutations, and the
# entry point into performing writes in our schema.
#
# This implements the following type system shorthand:
#   type Mutation {
#     introduceShip(input IntroduceShipInput!): IntroduceShipPayload
#   }
mutation_type = GraphQLObjectType(
    "Mutation", fields=lambda: {"introduceShip": ship_mutation}
)

# Finally, we construct our schema (whose starting query type is the query
# type we defined above) and export it.
star_wars_schema = GraphQLSchema(query=query_type, mutation=mutation_type)
